"""Time utilities."""

from datetime import datetime

from adaptive_oscillator.definitions import TIME_FORMAT


def time_str_to_seconds(time_str: str) -> float:
    """Convert a time string to seconds."""
    dt = datetime.strptime(time_str, TIME_FORMAT)
    return dt.hour * 3600 + dt.minute * 60 + dt.second + dt.microsecond / 1e6



"""Parser utils for log file data."""

from pathlib import Path

import numpy as np
import pandas as pd
from loguru import logger
from matplotlib import pyplot as plt

from adaptive_oscillator.base_classes import (
    AngleXYZ,
    Body,
    Joint,
    LeftRight,
    Limb,
    Quaternion,
    SensorFile,
    VectorXYZ,
)
from adaptive_oscillator.definitions import (
    ALPHA,
    ANGLES_SEGMENT_FIELDS,
    FIG_SIZE,
    IMU_SEGMENT_FIELDS,
    QUATERNION_SEGMENT_FIELDS,
    AnglesHeader,
    LogFileKeys,
    QuaternionHeader,
)
from adaptive_oscillator.utils.time_utils import time_str_to_seconds


class LogFiles:
    """Main entry point for accessing all sensor log files."""

    def __init__(self, base_path: str | Path) -> None:
        self._path = Path(base_path)
        self.accel = SensorFile(LogFileKeys.ACCEL, self._path)
        self.angle = SensorFile(LogFileKeys.ANGLE, self._path)
        self.gravity = SensorFile(LogFileKeys.GRAVITY, self._path)
        self.gyro = SensorFile(LogFileKeys.GYRO, self._path)
        self.quat = SensorFile(LogFileKeys.QUAT, self._path)

    def __repr__(self) -> str:  # pragma: no cover
        """Return a string representation of the LogFiles object."""
        return (
            f"Log files for dir: '{self._path}'"
            f"\n\t{self.accel.left}"
            f"\n\t{self.accel.right}"
            f"\n\t{self.angle.left}"
            f"\n\t{self.angle.right}"
            f"\n\t{self.gravity.left}"
            f"\n\t{self.gravity.right}"
            f"\n\t{self.gyro.left}"
            f"\n\t{self.gyro.right}"
            f"\n\t{self.quat.left}"
            f"\n\t{self.quat.right})"
        )

    def plot(self):
        """Plot log files."""
        logger.info("Plotting data.")
        accel_data = IMUParser(self.accel.right)
        accel_data.parse()
        accel_data.plot()

        gyro_data = IMUParser(self.gyro.right)
        gyro_data.parse()
        gyro_data.plot()

        quat_data = QuaternionParser(self.quat.right)
        quat_data.parse()
        quat_data.plot()

        accel_data = IMUParser(self.accel.left)
        accel_data.parse()
        accel_data.plot()

        gyro_data = IMUParser(self.gyro.left)
        gyro_data.parse()
        gyro_data.plot()

        quat_data = QuaternionParser(self.quat.left)
        quat_data.parse()
        quat_data.plot()


class IMUParser:
    """Parser for log files with limb information."""

    def __init__(self, filepath: Path):
        self.filepath = filepath
        self.time = np.array([])
        self.pelvis = VectorXYZ()
        self.upper_leg = VectorXYZ()
        self.lower_leg = VectorXYZ()
        self.foot = VectorXYZ()

    def parse(self):
        """Parse the log file and return a DataFrame."""
        raw_data = pd.read_csv(self.filepath, sep="\t+", engine="python")
        logger.debug(f"Parsing {self.filepath}")
        logger.debug(f"Columns: {raw_data.shape}")

        time_str = raw_data[AnglesHeader.TIME]
        self.time = np.array([time_str_to_seconds(t) for t in time_str])

        for segment_name, fields in IMU_SEGMENT_FIELDS.items():
            x = raw_data[fields[0]].to_numpy()
            y = raw_data[fields[1]].to_numpy()
            z = raw_data[fields[2]].to_numpy()
            setattr(self, segment_name, VectorXYZ(x, y, z))

    def plot(self):  # pragma: no cover
        """Plot the x, y, z data."""
        _, ax = plt.subplots(figsize=FIG_SIZE, sharex=True, nrows=4, ncols=1)

        for ii, (name, segment) in enumerate(
            zip(
                ["Pelvis", "Upper Leg", "Lower Leg", "Foot"],
                [self.pelvis, self.upper_leg, self.lower_leg, self.foot],
            )
        ):
            time = self.time - self.time[0]
            for axis in ["x", "y", "z"]:
                imu_signal = getattr(segment, axis)
                ax[ii].plot(time, imu_signal, label=f"{name}-{axis}", alpha=ALPHA)
            ax[ii].set_title(f"{name} - {self.filepath.stem}")
            ax[ii].set_xlabel("Time (s)")
            ax[ii].set_ylabel("Quaternion")
            ax[ii].legend(loc="upper right")
            ax[ii].grid(True)
            plt.tight_layout()


class AngleParser:
    """Parser for log files with angle."""

    def __init__(self, filepath: Path):
        self.filepath = filepath
        self.time = np.array([])
        self.hip = AngleXYZ()
        self.knee = AngleXYZ()
        self.ankle = AngleXYZ()

    def parse(self):
        """Parse the log file and return a DataFrame."""
        raw_data = pd.read_csv(self.filepath, sep="\t+", engine="python")
        logger.debug(f"Parsing {self.filepath}")
        logger.debug(f"Columns: {raw_data.shape}")

        time_str = raw_data[AnglesHeader.TIME]
        self.time = np.array([time_str_to_seconds(t) for t in time_str])

        for segment_name, fields in ANGLES_SEGMENT_FIELDS.items():
            x_deg = raw_data[fields[0]].to_numpy()
            y_deg = raw_data[fields[1]].to_numpy()
            z_deg = raw_data[fields[2]].to_numpy()
            setattr(self, segment_name, AngleXYZ(x_deg, y_deg, z_deg))


class QuaternionParser:
    """Parser for log files with quaternion information."""

    def __init__(self, filepath: Path):
        self.filepath = filepath
        self.time = np.array([])
        self.pelvis = Quaternion()
        self.upper_leg = Quaternion()
        self.lower_leg = Quaternion()
        self.foot = Quaternion()

    def parse(self):
        """Parse the log file and return a DataFrame."""
        raw_data = pd.read_csv(self.filepath, sep="\t")
        logger.debug(f"Parsing {self.filepath}")
        logger.debug(f"Columns: {raw_data.shape}")

        time_str = raw_data[QuaternionHeader.TIME]
        self.time = np.array([time_str_to_seconds(t) for t in time_str])

        for segment_name, fields in QUATERNION_SEGMENT_FIELDS.items():
            w = raw_data[fields[0]].to_numpy()
            x = raw_data[fields[1]].to_numpy()
            y = raw_data[fields[2]].to_numpy()
            z = raw_data[fields[3]].to_numpy()
            setattr(self, segment_name, Quaternion(w, x, y, z))

    def plot(self):  # pragma: no cover
        """Plot the Quaternion data."""
        _, ax = plt.subplots(figsize=FIG_SIZE, sharex=True, nrows=4, ncols=1)

        for ii, (name, segment) in enumerate(
            zip(
                ["Pelvis", "Upper Leg", "Lower Leg", "Foot"],
                [self.pelvis, self.upper_leg, self.lower_leg, self.foot],
            )
        ):
            time = self.time - self.time[0]
            for axis in ["w", "x", "y", "z"]:
                quat_component = getattr(segment, axis)
                ax[ii].plot(time, quat_component, label=f"{name}-{axis}", alpha=ALPHA)
            ax[ii].set_title(f"{name} Orientation")
            ax[ii].set_xlabel("Time (s)")
            ax[ii].set_ylabel("Quaternion")
            ax[ii].legend(loc="upper right")
            ax[ii].grid(True)
            plt.tight_layout()


class LogParser:
    """Parser for log files with limb information."""

    def __init__(self, log_files: LogFiles):
        logger.info(f"Parsing {log_files}")
        accel_data_right = IMUParser(log_files.accel.right)
        accel_data_right.parse()
        accel_data_left = IMUParser(log_files.accel.left)
        accel_data_left.parse()

        gyro_data_right = IMUParser(log_files.gyro.right)
        gyro_data_right.parse()
        gyro_data_left = IMUParser(log_files.gyro.left)
        gyro_data_left.parse()

        quat_data_right = QuaternionParser(log_files.quat.right)
        quat_data_right.parse()
        quat_data_left = QuaternionParser(log_files.quat.left)
        quat_data_left.parse()

        angles_right = AngleParser(log_files.angle.right)
        angles_right.parse()
        angles_left = AngleParser(log_files.angle.left)
        angles_left.parse()

        time = accel_data_right.time

        pelvis_right = Limb(
            time=time,
            accel=accel_data_right.pelvis,
            gyro=gyro_data_right.pelvis,
            quat=quat_data_right.pelvis,
        )
        upper_leg_right = Limb(
            time=time,
            accel=accel_data_right.upper_leg,
            gyro=gyro_data_right.upper_leg,
            quat=quat_data_right.upper_leg,
        )
        lower_leg_right = Limb(
            time=time,
            accel=accel_data_right.lower_leg,
            gyro=gyro_data_right.lower_leg,
            quat=quat_data_right.lower_leg,
        )
        foot_right = Limb(
            time=time,
            accel=accel_data_right.foot,
            gyro=gyro_data_right.foot,
            quat=quat_data_right.foot,
        )
        hip_right = Joint(time=time, angles=angles_right.hip)
        knee_right = Joint(time=time, angles=angles_right.knee)
        ankle_right = Joint(time=time, angles=angles_right.ankle)

        pelvis_left = Limb(
            time=time,
            accel=accel_data_left.pelvis,
            gyro=gyro_data_left.pelvis,
            quat=quat_data_left.pelvis,
        )
        upper_leg_left = Limb(
            time=time,
            accel=accel_data_left.upper_leg,
            gyro=gyro_data_left.upper_leg,
            quat=quat_data_left.upper_leg,
        )
        lower_leg_left = Limb(
            time=time,
            accel=accel_data_left.lower_leg,
            gyro=gyro_data_left.lower_leg,
            quat=quat_data_left.lower_leg,
        )
        foot_left = Limb(
            time=time,
            accel=accel_data_left.foot,
            gyro=gyro_data_left.foot,
            quat=quat_data_left.foot,
        )
        hip_left = Joint(time=time, angles=angles_left.hip)
        knee_left = Joint(time=time, angles=angles_left.knee)
        ankle_left = Joint(time=time, angles=angles_left.ankle)

        self.log_files = log_files
        self.time = accel_data_right.time
        self.data = LeftRight(
            left=Body(
                pelvis=pelvis_left,
                upper_leg=upper_leg_left,
                lower_leg=lower_leg_left,
                foot=foot_left,
                hip=hip_left,
                knee=knee_left,
                ankle=ankle_left,
            ),
            right=Body(
                pelvis=pelvis_right,
                upper_leg=upper_leg_right,
                lower_leg=lower_leg_right,
                foot=foot_right,
                hip=hip_right,
                knee=knee_right,
                ankle=ankle_right,
            ),
        )



"""Common base classes."""

from dataclasses import dataclass, field
from pathlib import Path

import numpy as np
from numpy.typing import NDArray

from adaptive_oscillator.definitions import LOG_FILE_EXT


@dataclass
class VectorXYZ:
    """XYZ Vector."""

    x: NDArray = field(default_factory=lambda: np.array([]))
    y: NDArray = field(default_factory=lambda: np.array([]))
    z: NDArray = field(default_factory=lambda: np.array([]))

    def __getitem__(self, index: int | slice) -> NDArray:
        """Return stacked XYZ components as a 2D array, or select one component.

        :param index: Index or slice for accessing stacked vector components.
                      0 = x, 1 = y, 2 = z; or a slice like 0:2
        :return: Stacked NumPy array of shape (3, N) or (K, N).
        :raises IndexError: If index is out of bounds.
        """
        stacked = np.stack(arrays=[self.x, self.y, self.z], axis=1)
        return stacked[index].T

    def __len__(self) -> int:
        """Return the number of elements in the vector."""
        return len(self.x)


@dataclass
class Quaternion:
    """Quaternion."""

    w: NDArray = field(default_factory=lambda: np.array([]))
    x: NDArray = field(default_factory=lambda: np.array([]))
    y: NDArray = field(default_factory=lambda: np.array([]))
    z: NDArray = field(default_factory=lambda: np.array([]))

    def __getitem__(self, index: int | slice) -> NDArray:
        """Return stacked XYZ components as a 2D array, or select one component.

        :param index: Index or slice for accessing stacked vector components.
                      0 = x, 1 = y, 2 = z; or a slice like 0:2
        :return: Stacked NumPy array of shape (3, N) or (K, N).
        :raises IndexError: If index is out of bounds.
        """
        stacked = np.stack([self.w, self.x, self.y, self.z], axis=1)
        return stacked[index].T

    def __mul__(self, quat_b: "Quaternion") -> "Quaternion":
        """Multiply two quaternions.

        :param quat_b: Quaternion to multiply with.
        :return: New Quaternion representing the product.
        """
        q = quat_b
        w = self.w * q.w - self.x * q.x - self.y * q.y - self.z * q.z
        x = self.w * q.x + self.x * q.w + self.y * q.z - self.z * q.y
        y = self.w * q.y - self.x * q.z + self.y * q.w + self.z * q.x
        z = self.w * q.z + self.x * q.y - self.y * q.x + self.z * q.w
        return Quaternion(w, x, y, z)


@dataclass
class AngleXYZ:
    """XYZ Angle Vector."""

    x_deg: NDArray = field(default_factory=lambda: np.array([]))
    y_deg: NDArray = field(default_factory=lambda: np.array([]))
    z_deg: NDArray = field(default_factory=lambda: np.array([]))

    def __getitem__(self, index: int | slice) -> NDArray:
        """Return stacked XYZ components as a 2D array, or select one component.

        :param index: Index or slice for accessing stacked vector components.
                      0 = x, 1 = y, 2 = z; or a slice like 0:2
        :return: Stacked NumPy array of shape (3, N) or (K, N).
        :raises IndexError: If index is out of bounds.
        """
        stacked = np.stack([self.x_deg, self.y_deg, self.z_deg], axis=1)
        return stacked[index].T

    def __len__(self) -> int:
        """Return the number of elements in the vector."""
        return len(self.x_deg)


class SensorFile:
    """Represent a sensor category with left and right side access."""

    def __init__(self, category: str, base_path: Path) -> None:
        self.left = base_path / f"{category}_left{LOG_FILE_EXT}"
        self.right = base_path / f"{category}_right{LOG_FILE_EXT}"


@dataclass
class Limb:
    """Represent a side of the body."""

    time: NDArray
    accel: VectorXYZ
    gyro: VectorXYZ
    quat: Quaternion


@dataclass
class Joint:
    """Represent a side of the body."""

    time: NDArray
    angles: AngleXYZ


@dataclass
class Body:
    """Represent a body."""

    pelvis: Limb
    upper_leg: Limb
    lower_leg: Limb
    foot: Limb
    hip: Joint
    knee: Joint
    ankle: Joint


@dataclass
class LeftRight:
    """Represent a body."""

    left: Body
    right: Body



"""Adaptive Oscillator gait tracking."""

from dataclasses import dataclass

import numpy as np
from loguru import logger
from numpy.typing import NDArray
from scipy.integrate import solve_ivp
from scipy.interpolate import CubicSpline

from adaptive_oscillator.definitions import ETA, N_HARMONICS, NU_OMEGA, NU_PHI


# -----------------------------------------------------------------------------
# Parameters
# -----------------------------------------------------------------------------
@dataclass
class AOParameters:
    """Adaptive Oscillator parameters."""

    eta: float = ETA
    nu_phi: float = NU_PHI
    nu_omega: float = NU_OMEGA
    n_harmonics: int = N_HARMONICS


# -----------------------------------------------------------------------------
# Adaptive Oscillator
# -----------------------------------------------------------------------------
class AdaptiveOscillator:
    """Adaptive Oscillator tracking rhythmic signals like inter-limb hip angle."""

    def __init__(self, params: AOParameters, omega_init: float = 1.0):
        self.params = params
        self.n = params.n_harmonics
        self.omega = omega_init
        self.alpha_0 = 0.0
        self.alpha = np.zeros(self.n)
        self.phi = np.zeros(self.n)
        self.last_t = 0.0
        self.theta_hat = 0.0

    def _dynamics(self, t: float, y: NDArray, theta_il: float) -> NDArray:
        omega = y[0]
        alpha_0 = y[1]
        alpha = y[2 : 2 + self.n]
        phi = y[2 + self.n :]

        theta_hat = alpha_0 + np.sum(alpha * np.sin(phi))
        F = theta_il - theta_hat
        alpha_sum = np.sum(alpha) + 1e-6

        phi_dot = (
            omega * np.arange(1, self.n + 1)
            + self.params.nu_phi * F * np.cos(phi) / alpha_sum
        )
        omega_dot = self.params.nu_omega * F * np.cos(phi[0]) / alpha_sum
        dalpha = self.params.eta * F * np.sin(phi)
        dalpha_0 = self.params.eta * F

        return np.concatenate([[omega_dot, dalpha_0], dalpha, phi_dot])

    def update(self, t: float, theta_il: float, solver: str = "RK45") -> float:
        """Integrate the oscillator from self.last_t to t, return gait phase φ_GP(t)."""
        y0 = np.concatenate([[self.omega, self.alpha_0], self.alpha, self.phi])
        sol = solve_ivp(
            fun=lambda t_, y_: self._dynamics(t_, y_, theta_il),
            t_span=(self.last_t, t),
            y0=y0,
            method=solver,
            max_step=0.01,
        )

        y = sol.y[:, -1]
        self.omega = y[0]
        self.alpha_0 = y[1]
        self.alpha = y[2 : 2 + self.n]
        self.phi = y[2 + self.n :]
        self.last_t = t

        self.theta_hat = self.alpha_0 + np.sum(self.alpha * np.sin(self.phi))
        return np.mod(self.phi[0], 2 * np.pi)


# -----------------------------------------------------------------------------
# Gait Phase Estimation
# -----------------------------------------------------------------------------
class GaitPhaseEstimator:
    """Estimates corrected gait phase using AOs, event detection, and correction."""

    def __init__(self, params: AOParameters):
        self.ao = AdaptiveOscillator(params)
        self.last_t_start = -np.inf
        self.phi_error = 0.0
        self.ke = 1.0
        self.phi_gp = 0.0

    def detect_gait_event(
        self, t: float, theta_il: float, theta_il_dot: float, period: float
    ) -> bool:
        """Detect gait event based on theta_il and theta_il_dot."""
        return theta_il_dot < 0.01 and t - self.last_t_start > 0.7 * period

    def correct_phase(
        self, phi_gp: float, t: float, t_start: float, omega: float
    ) -> float:
        """Correct gait phase φ(t) using error correction."""
        Pe = -phi_gp if 0 <= phi_gp < np.pi else 2 * np.pi - phi_gp
        Ce = self.ke * (Pe - self.phi_error)
        self.phi_error += Ce * np.exp(-omega * (t - t_start))
        return np.mod(phi_gp + self.phi_error, 2 * np.pi)

    def update(self, t: float, theta_il: float, theta_il_dot: float) -> float:
        """Update gait phase and return corrected gait phase φ(t)."""
        self.phi_gp = self.ao.update(t, theta_il)
        omega = self.ao.omega
        period = 2 * np.pi / omega

        if self.detect_gait_event(t, theta_il, theta_il_dot, period):
            self.last_t_start = t

        phi = self.correct_phase(self.phi_gp, t, self.last_t_start, omega)

        logger.debug(
            f"t={t:.2f}, φ_GP={self.phi_gp:.2f}, "
            f"φ={phi:.2f}, ω={omega:.2f}, θ_hat={self.ao.theta_hat:.2f}"
        )
        return phi


# -----------------------------------------------------------------------------
# PID Controller
# -----------------------------------------------------------------------------
class PIDController:
    """PID controller."""

    def __init__(self, kp: float, ki: float, kd: float):
        self.kp = kp
        self.ki = ki
        self.kd = kd
        self.integral = 0.0
        self.last_error = 0.0

    def compute(self, error: float, dt: float) -> float:
        """Compute PID output."""
        self.integral += error * dt
        derivative = (error - self.last_error) / dt
        self.last_error = error
        return self.kp * error + self.ki * self.integral + self.kd * derivative


# -----------------------------------------------------------------------------
# Low-Level Motor Controller
# -----------------------------------------------------------------------------
class LowLevelController:
    """Low-level motor controller."""

    def __init__(
        self,
        kp: float = 5.0,
        ki: float = 0.0,
        kd: float = 0.1,
        gait_shape: NDArray | None = None,
    ):
        self.pid = PIDController(kp, ki, kd)
        x = np.linspace(0, 2 * np.pi, 100)
        y = gait_shape if gait_shape is not None else np.sin(x)
        self.spline = CubicSpline(x, y)

    def compute(self, phi: float, theta_m: float, dt: float) -> float:
        """Compute motor output."""
        theta_r = self.spline(phi - np.pi)
        error = theta_r - theta_m
        return self.pid.compute(error, dt)  # type: ignore[arg-type]



"""Controller module for the Adaptive Oscillator."""

import time

from loguru import logger

from adaptive_oscillator.definitions import DEFAULT_DELTA_TIME
from adaptive_oscillator.oscillator import (
    AOParameters,
    GaitPhaseEstimator,
    LowLevelController,
)
from adaptive_oscillator.utils.plot_utils import RealtimeAOPlotter


class AOController:
    """Encapsulate the AO control loop and optional real-time plotting."""

    def __init__(
        self,
        config: AOParameters | None = None,
        show_plots: bool = False,
        ssh: bool = False,
    ):
        """Initialize controller.

        :param show_plots: Plot IMU logs before running the control loop.
        """
        self.params = AOParameters() if config is None else config
        self.estimator = GaitPhaseEstimator(self.params)
        self.controller = LowLevelController()
        self.theta_m = 0.0
        self.last_time: float | None = None

        self.ang_idx = 0

        self.motor_output: list[float] = []
        self.theta_hat_output: list[float] = []
        self.phi_gp_output: list[float] = []
        self.omegas: list[float] = []

        self.plotter: RealtimeAOPlotter | None = None
        if show_plots:  # pragma: no cover
            self.plotter = RealtimeAOPlotter(ssh=ssh)
            self.plotter.run()

    def step(self, t: float, th: float, dth: float) -> tuple[float, float, float]:
        """Step the AO ahead with one frame of data from the IMU."""
        if self.last_time is None:
            dt = DEFAULT_DELTA_TIME
        else:
            dt = t - self.last_time
        self.last_time = t

        phi = self.estimator.update(t=t, theta_il=th, theta_il_dot=dth)
        omega_cmd = self.controller.compute(phi=phi, theta_m=self.theta_m, dt=dt)
        self.theta_m += omega_cmd * dt

        # Store outputs
        self.motor_output.append(self.theta_m)
        self.theta_hat_output.append(self.estimator.ao.theta_hat)
        self.phi_gp_output.append(self.estimator.phi_gp)
        self.omegas.append(self.estimator.ao.omega)

        theta_hat = self.estimator.ao.theta_hat
        omega = self.estimator.ao.omega
        phi_gp = self.estimator.phi_gp
        logger.info(
            f"theta_hat: {theta_hat:.2f}, omega: {omega:.2f}, phi_gp: {phi_gp:.2f}"
        )

        # Update live plot if enabled
        if self.plotter is not None:  # pragma: no cover
            self.plotter.update_data(
                t=t,
                theta_il=th,
                theta_hat=self.estimator.ao.theta_hat,
                omega=self.estimator.ao.omega,
                phi_gp=self.estimator.phi_gp,
            )
            time.sleep(dt)

        return theta_hat, omega, phi_gp



"""Common definitions for my module."""

import sys

import numpy as np
from loguru import logger

# plot definitions
FIG_SIZE = (14, 10)  # width=8 inches, height=6 inches
ALPHA = 0.8

TIME_FORMAT = "%H:%M:%S.%f"

LOG_FILE_EXT = ".txt"
logger.configure(handlers=[{"sink": sys.stderr, "level": "INFO"}])

NUMPY_PRINT_PRECISION = 3
np.set_printoptions(precision=NUMPY_PRINT_PRECISION)

# AOParameters
ETA = 0.05
N_HARMONICS = 3
NU_PHI = 0.5
NU_OMEGA = 0.5

DEFAULT_DELTA_TIME = 0.01


class LogFileKeys:
    """Enum for the log file categories."""

    ACCEL = "Accelerometers"
    ANGLE = "Angles"
    GRAVITY = "Gravity"
    GYRO = "Gyroscopes"
    QUAT = "Quaternions"


class IMUHeader:
    """Headers for the IMU sensor data."""

    TIME = "Time"
    PELVIS_X = "Pelvis_x"
    PELVIS_Y = "Pelvis_y"
    PELVIS_Z = "Pelvis_z"
    UPPLEG_X = "UppLeg_x"
    UPPLEG_Y = "UppLeg_y"
    UPPLEG_Z = "UppLeg_z"
    LOWLEG_X = "LowLeg_x"
    LOWLEG_Y = "LowLeg_y"
    LOWLEG_Z = "LowLeg_z"
    FOOT_X = "Foot_x"
    FOOT_Y = "Foot_y"
    FOOT_Z = "Foot_z"


class QuaternionHeader:
    """Headers for the quaternion data."""

    TIME = "Time"
    PELVIS_W = "Pelvis_w"
    PELVIS_X = "Pelvis_x"
    PELVIS_Y = "Pelvis_y"
    PELVIS_Z = "Pelvis_z"
    UPPLEG_W = "UppLeg_w"
    UPPLEG_X = "UppLeg_x"
    UPPLEG_Y = "UppLeg_y"
    UPPLEG_Z = "UppLeg_z"
    LOWLEG_W = "LowLeg_w"
    LOWLEG_X = "LowLeg_x"
    LOWLEG_Y = "LowLeg_y"
    LOWLEG_Z = "LowLeg_z"
    FOOT_W = "Foot_w"
    FOOT_X = "Foot_x"
    FOOT_Y = "Foot_y"
    FOOT_Z = "Foot_z"


class AnglesHeader:
    """Headers for the angle data."""

    TIME = "Time"
    HIP_X = "Hip_x"
    HIP_Y = "Hip_y"
    HIP_Z = "Hip_z"
    KNEE_X = "Knee_x"
    KNEE_Y = "Knee_y"
    KNEE_Z = "Knee_z"
    ANKLE_X = "Ankle_x"
    ANKLE_Y = "Ankle_y"
    ANKLE_Z = "Ankle_z"


IMU_SEGMENT_FIELDS = {
    "pelvis": [
        IMUHeader.PELVIS_X,
        IMUHeader.PELVIS_Y,
        IMUHeader.PELVIS_Z,
    ],
    "upper_leg": [
        IMUHeader.UPPLEG_X,
        IMUHeader.UPPLEG_Y,
        IMUHeader.UPPLEG_Z,
    ],
    "lower_leg": [
        IMUHeader.LOWLEG_X,
        IMUHeader.LOWLEG_Y,
        IMUHeader.LOWLEG_Z,
    ],
    "foot": [
        IMUHeader.FOOT_X,
        IMUHeader.FOOT_Y,
        IMUHeader.FOOT_Z,
    ],
}

QUATERNION_SEGMENT_FIELDS = {
    "pelvis": [
        QuaternionHeader.PELVIS_W,
        QuaternionHeader.PELVIS_X,
        QuaternionHeader.PELVIS_Y,
        QuaternionHeader.PELVIS_Z,
    ],
    "upper_leg": [
        QuaternionHeader.UPPLEG_W,
        QuaternionHeader.UPPLEG_X,
        QuaternionHeader.UPPLEG_Y,
        QuaternionHeader.UPPLEG_Z,
    ],
    "lower_leg": [
        QuaternionHeader.LOWLEG_W,
        QuaternionHeader.LOWLEG_X,
        QuaternionHeader.LOWLEG_Y,
        QuaternionHeader.LOWLEG_Z,
    ],
    "foot": [
        QuaternionHeader.FOOT_W,
        QuaternionHeader.FOOT_X,
        QuaternionHeader.FOOT_Y,
        QuaternionHeader.FOOT_Z,
    ],
}

ANGLES_SEGMENT_FIELDS = {
    "hip": [
        AnglesHeader.HIP_X,
        AnglesHeader.HIP_Y,
        AnglesHeader.HIP_Z,
    ],
    "knee": [
        AnglesHeader.KNEE_X,
        AnglesHeader.KNEE_Y,
        AnglesHeader.KNEE_Z,
    ],
    "ankle": [
        AnglesHeader.ANKLE_X,
        AnglesHeader.ANKLE_Y,
        AnglesHeader.ANKLE_Z,
    ],
}



"""Plot utilities module."""

import logging
import threading
from collections import deque
from dataclasses import dataclass

import plotly.graph_objects as go
from dash import Dash, dcc, html
from dash.dependencies import Input, Output

log = logging.getLogger("werkzeug")
log.setLevel(logging.ERROR)


@dataclass
class PlotMetrics:
    """AO plot metrics."""

    time_data = "time_data"
    theta_hat = "theta_hat"
    theta_il = "theta_il"
    omega = "omega"
    phi_gp = "phi_gp"


class RealtimeAOPlotter:  # pragma: no cover
    """Real-time Data plotter for the Adaptive Oscillator."""

    def __init__(
        self, ssh: bool = False, window_sec: float = 5.0, frequency_hz: int = 100
    ) -> None:
        self.window_sec = window_sec
        self.data_points = int(window_sec * frequency_hz)

        self.app = Dash(__name__)
        self._setup_layout()
        self._register_callbacks()

        # Thread-safe buffers
        self._lock = threading.Lock()
        self.data: dict = {
            PlotMetrics.time_data: deque(maxlen=self.data_points),
            PlotMetrics.theta_il: deque(maxlen=self.data_points),
            PlotMetrics.theta_hat: deque(maxlen=self.data_points),
            PlotMetrics.omega: deque(maxlen=self.data_points),
            PlotMetrics.phi_gp: deque(maxlen=self.data_points),
        }

        self.host = "0.0.0.0" if ssh else "127.0.0.1"  # ruff: ignore B104,S104
        self.port = 8050

    def _setup_layout(self) -> None:
        self.app.layout = html.Div(
            [
                html.H1(
                    "Adaptive Oscillator Logfile Replay",
                    style={"margin-bottom": "10px"},
                ),
                dcc.Graph(
                    id="hip-angle-graph", style={"height": "29vh", "margin": "0"}
                ),
                dcc.Graph(
                    id="omega-estimate-graph", style={"height": "29vh", "margin": "0"}
                ),
                dcc.Graph(
                    id="gait-phase-graph", style={"height": "29vh", "margin": "0"}
                ),
                dcc.Interval(
                    id="interval-component", interval=200, n_intervals=0
                ),  # 5 Hz is plenty over SSH
            ],
            style={"padding": "10px", "gap": "0px"},
        )

    def _register_callbacks(self) -> None:
        @self.app.callback(
            [
                Output("hip-angle-graph", "figure"),
                Output("omega-estimate-graph", "figure"),
                Output("gait-phase-graph", "figure"),
            ],
            Input("interval-component", "n_intervals"),
        )
        def update_graphs(_):
            return self._generate_figures()

    def _snapshot(
        self,
    ) -> tuple[list[float], list[float], list[float], list[float], list[float]]:
        """Copy current data under a lock to avoid torn reads."""
        with self._lock:
            time_data = list(self.data[PlotMetrics.time_data])
            theta_il = list(self.data[PlotMetrics.theta_il])
            theta_hat = list(self.data[PlotMetrics.theta_hat])
            omega = list(self.data[PlotMetrics.omega])
            phi_gp = list(self.data[PlotMetrics.phi_gp])
        return time_data, theta_il, theta_hat, omega, phi_gp

    def _generate_figures(self) -> tuple[go.Figure, go.Figure, go.Figure]:
        time_data, theta_il, theta_hat, omega, phi_gp = self._snapshot()
        if not time_data:
            empty = go.Figure()
            return empty, empty, empty

        # Keep last window
        latest = time_data[-1]
        window_start = latest - self.window_sec
        start_idx = next((i for i, t in enumerate(time_data) if t >= window_start), 0)
        time_data, theta_il, theta_hat, omega, phi_gp = (
            time_data[start_idx:],
            theta_il[start_idx:],
            theta_hat[start_idx:],
            omega[start_idx:],
            phi_gp[start_idx:],
        )

        margin = dict(l=30, r=10, t=30, b=30)

        hip_fig = go.Figure()
        hip_fig.add_trace(
            go.Scatter(x=time_data, y=theta_il, mode="lines", name="θ_IL (input)")
        )
        hip_fig.add_trace(
            go.Scatter(x=time_data, y=theta_hat, mode="lines", name="θ̂ (estimated)")
        )
        hip_fig.update_layout(
            title="Input vs Estimated Hip Angle",
            xaxis_title="Time (s)",
            yaxis_title="Angle (rad)",
            margin=margin,
            legend=dict(
                orientation="h", yanchor="bottom", y=1.02, xanchor="right", x=1
            ),
        )

        omega_fig = go.Figure()
        omega_fig.add_trace(
            go.Scatter(
                x=time_data, y=omega, mode="lines", name="ω", line=dict(color="green")
            )
        )
        omega_fig.update_layout(
            title="Omega Estimate",
            xaxis_title="Time (s)",
            yaxis_title="Angle (rad)",
            margin=margin,
            legend=dict(
                orientation="h", yanchor="bottom", y=1.02, xanchor="right", x=1
            ),
        )

        phase_fig = go.Figure()
        phase_fig.add_trace(
            go.Scatter(
                x=time_data,
                y=phi_gp,
                mode="lines",
                name="ϕ_GP",
                line=dict(color="purple"),
            )
        )
        phase_fig.update_layout(
            title="Estimated Gait Phase",
            xaxis_title="Time (s)",
            yaxis_title="Phase (rad)",
            margin=margin,
            legend=dict(
                orientation="h", yanchor="bottom", y=1.02, xanchor="right", x=1
            ),
        )

        return hip_fig, omega_fig, phase_fig

    def update_data(
        self, t: float, theta_il: float, theta_hat: float, omega: float, phi_gp: float
    ) -> None:
        """Update the data and plot it."""
        with self._lock:
            self.data[PlotMetrics.time_data].append(t)
            self.data[PlotMetrics.theta_il].append(theta_il)
            self.data[PlotMetrics.theta_hat].append(theta_hat)
            self.data[PlotMetrics.omega].append(omega)
            self.data[PlotMetrics.phi_gp].append(phi_gp)

    def run(self, threaded: bool = True) -> None:
        """Run the AO control loop."""
        if threaded:
            th = threading.Thread(
                target=self.app.run,
                kwargs={
                    "debug": False,
                    "use_reloader": False,
                    "host": self.host,
                    "port": self.port,
                },
                daemon=False,  # keep process alive
            )
            th.start()
        else:
            self.app.run_server(
                debug=False, use_reloader=False, host=self.host, port=self.port
            )
